/*
expr.h
Copyright © 2008 William Astle

This file is part of LWASM.

LWASM is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*
Definitions for expression evaluator
*/

#ifndef __expr_h_seen__
#define __expr_h_seen__

#ifndef __expr_c_seen__
#define __expr_E__ extern
#else
#define __expr_E__
#endif

// term types
#define LWASM_TERM_NONE		0
#define LWASM_TERM_OPER		1	// an operator
#define LWASM_TERM_INT		2	// 32 bit signed integer
#define LWASM_TERM_SYM		3	// symbol reference
#define LWASM_TERM_SECBASE	4	// section base reference

// operator types
#define LWASM_OPER_NONE		0
#define LWASM_OPER_PLUS		1	// +
#define LWASM_OPER_MINUS	2	// -
#define LWASM_OPER_TIMES	3	// *
#define LWASM_OPER_DIVIDE	4	// /
#define LWASM_OPER_MOD		5	// %
#define LWASM_OPER_INTDIV	6	// \ (don't end line with \)
#define LWASM_OPER_BWAND	7	// bitwise AND
#define LWASM_OPER_BWOR		8	// bitwise OR
#define LWASM_OPER_BWXOR	9	// bitwise XOR
#define LWASM_OPER_AND		10	// boolean AND
#define LWASM_OPER_OR		11	// boolean OR
#define LWASM_OPER_NEG		12	// - unary negation (2's complement)
#define LWASM_OPER_COM		13	// ^ unary 1's complement


// term structure
typedef struct lwasm_expr_term_s
{
	int term_type;		// type of term (see above)
	char *symbol;		// name of a symbol
	int value;			// value of the term (int) or operator number (OPER)
} lwasm_expr_term_t;

// type for an expression evaluation stack
typedef struct lwasm_expr_stack_node_s lwasm_expr_stack_node_t;
struct lwasm_expr_stack_node_s
{
	lwasm_expr_term_t		*term;
	lwasm_expr_stack_node_t	*prev;
	lwasm_expr_stack_node_t	*next;	
};

typedef struct lwasm_expr_stack_s
{
	lwasm_expr_stack_node_t *head;
	lwasm_expr_stack_node_t *tail;
} lwasm_expr_stack_t;

__expr_E__ void lwasm_expr_term_free(lwasm_expr_term_t *t);
__expr_E__ lwasm_expr_term_t *lwasm_expr_term_create_oper(int oper);
__expr_E__ lwasm_expr_term_t *lwasm_expr_term_create_sym(char *sym);
__expr_E__ lwasm_expr_term_t *lwasm_expr_term_create_int(int val);
__expr_E__ lwasm_expr_term_t *lwasm_expr_term_create_secbase(void);
__expr_E__ lwasm_expr_term_t *lwasm_expr_term_dup(lwasm_expr_term_t *t);

__expr_E__ void lwasm_expr_stack_free(lwasm_expr_stack_t *s);
__expr_E__ lwasm_expr_stack_t *lwasm_expr_stack_create(void);

__expr_E__ void lwasm_expr_stack_push(lwasm_expr_stack_t *s, lwasm_expr_term_t *t);
__expr_E__ lwasm_expr_term_t *lwasm_expr_stack_pop(lwasm_expr_stack_t *s);

/*
Evaluate an expression. The result is an lwasm_expr_stack_t pointer. If the
expression evaluates to a constant result, the stack will contain exactly one
value which will be a constant. Otherwise, the stack will contain the
expression with all operations that can be evaluated completely evaluated.
You must call lwasm_expr_stack_free() on the result when you are finished
with it.
*/
__expr_E__ lwasm_expr_stack_t *lwasm_expr_eval(const char *inp, const char **outp, lwasm_expr_stack_t *(*sfunc)(char *sym, void *state), void *state);

// simplify expression
__expr_E__ int lwasm_expr_reval(lwasm_expr_stack_t *s, lwasm_expr_stack_t *(*sfunc)(char *sym, void *state), void *state);

// useful macros
// is the expression "simple" (one term)?
#define lwasm_expr_is_simple(s) ((s) -> head == (s) -> tail)

// is the expression constant?
#define lwasm_expr_is_constant(s) (lwasm_expr_is_simple(s) && (!((s) -> head) || (s) -> head -> term -> term_type == LWASM_TERM_INT))

// get the constant value of an expression or 0 if not constant
#define lwasm_expr_get_value(s) (lwasm_expr_is_constant(s) ? ((s) -> head ? (s) -> head -> term -> value : 0) : 0)

#undef __expr_E__

#endif // __expr_h_seen__
